/* Retro-RPN calculator */
/* CC BY-NC-SA Jeroen Brinkman */
/* Release 3: dd 16-10-2024 */

// R-2 fixes faulty DST routine
// R-3 fixes problem with deleting DP
#define SOFTWARE  "R-3" // Software release

#include <Arduino.h>
#include <EEPROM.h>
#include <TM1637TinyDisplay6.h> // https://github.com/jasonacox/TM1637TinyDisplay
#include <fp64lib.h>            // https://fp64lib.org/download/
#include <RTClib.h>             // https://github.com/adafruit/rtclib 

RTC_DS3231 rtc;

// Feel free to adjust to your own taste & requirements if you know what you are doing
#define TEST     false // true or false If true, use the serial monitor to show debugging info. 
                       // CAREFULLY: Only set true when Arduino NANO Every is used, normal NANO will not fit!
#define BAUD    115200 // Baud rate for the serial monitor
#define USEDST    true // true or false Use daylight saving time (true) or not (false)
#define DUTCH     true // true or false Set to true for Dutch, false for English
#define SLEEP    60000 // Time in milliseconds before going to sleep
#define SEEDELAY  2000 // Time to wait in milliseconds to check the correctness of input
#define STACK_SIZE   8 // Size of stack

// Predefined memory values adjust to your own taste & requirements
#define MEM0     1.852 // Initial value in memory 1         - seamile to km
#define MEM1   1.60934 // Initial value in memory 1         - mile to km
#define MEM2      2.54 // Initial value in memory 2         - inches to cm
#define MEM3 float64_NUMBER_PI // Initial value in memory 3 - pi
#define MEM4     30.48 // Initial value in memory 4         - feet to cm
#define MEM5   14.5038 // Initial value in memory 5         - psi to bar
#define MEM6         0 // Not used
#define MEM7         0 // Not used
#define MEM8         0 // Not used
#define MEM9         0 // Not used

// Preferable use the following I/O ports
#define UIT1       2   // Set Left  column 1 {yellow wire}
#define UIT2       3   // Set       Column 2 {orange wire}
#define UIT3       4   // Set       Column 3 {red wire}
#define LUP        5   // Upper display entry LED - Must be PWM port
#define LLO        6   // Lower display entry LED - Must be PWM port
#define UIT4       7   // Set Right column 4 {brown wire}
#define IN1        8   // Read  Bottom row 1 {green wire}
#define IN2        9   // Read         row 2 {light blue wire}
#define IN3       10   // Read         row 3 {blue wire}
#define IN4       11   // Read         row 4 {brown-purple wire}
#define IN5       12   // Read     Top row 5 {purple wire}
#define HEART     13   // On-board heartbeat
#define CUP       A0   // Upper display CLK  {orange wire}
#define DUP       A1   // Upper display DIO  {brown wire}
#define CLO       A2   // Lower display CLK  {light blue wire} 
#define DLO       A3   // Lower display DIO  {dark blue wire} 
//      SDA       A4      I2C Data  - Must be I2C Data
//      SDL       A5      I2C Clock - Must be I2C Clock

//Please leave it as it is
#define BRICOR       0 // Brightness correction for lower display to match brightness both displays (if necessary)
#define MINPWM      10 // Minimal brightness for PWM LED's (0-255)
#define MAXPWM     100 // Maximal brightness for PWM LED's (0-255)
#define DEBOUNCE    25 // Number of debounce samples
#define WAIT       150 // Number of milliseconds to wait after pressing the enter key (extra debouncing)

#define STACK        1 // There is a number on the stack
#define CURSOR     '_' // String underscore

#define UPLEN        6 // Maximum number of digits that can be displayed on the upper display   
#define LOLEN        6 // Maximum number of digits that can be displayed on the lower display     
#define MAXLEN      12 // Maximum number of digits that may be entered  = UPLEN + LOLEN
#define STRING_LEN  14 // input string length = MAXLEN + 2

#define MODUS_SLEEP  0 // Calculator sleeps
#define MODUS_CLOCK  1 // Calculator shows time
#define MODUS_CALC   2 // Calculator performs calculations
#define MODUS_COUNT  3 // Calculator shows count down

#define SET_OFF      1 // No time or count down is shown, so go to sleep
#define SET_SCLOCK   2 // Calculator shows time and goes to sleep
#define SET_SCOUNT   3 // Calculator shows count down and goes to sleep
#define SET_PCLOCK   4 // Calculator shows time permanent
#define SET_PCOUNT   5 // Calculator shows count down permanent

#define STATE_WAIT   3 // Calculator waits for input
#define STATE_INT    4 // No decimal point has been entered so far
#define STATE_FRAC   5 // Number input is behind the decimal point

#define SHOW_ENG     1 // x.x | Ey
#define SHOW_TWO     2 // x.x | y.y
#define SHOW_EXT     3 // xxx | x.x 
#define SHOW_SPLIT   4 //   x.| x 

#define DD           1 // First the day - > change if required
#define MM           2 // Then the month - > change if required
#define YY           3 // finally the year - > change if required

#define SELECT       0 // select line for keyboard scan 
#define DESELECT     1 // deselect line for keyboard scan 

#define DAY          1 // EEPROM position for storage count-down year
#define MONTH        2 // EEPROM position for storage count-down month
#define YEAR         3 // EEPROM position for storage count-down day
#define HOUR         4 // EEPROM position for storage count-down hour
#define MINUTE       5 // EEPROM position for storage count-down minute
#define SECOND       6 // EEPROM position for storage count-down minute
#define SET          7 // EEPROM position for storage set setting
#define FIX          8 // EEPROM position for storage fix setting
#define NOFIX        9 // EEPROM position for storage nofix setting
#define SPLIT       10 // EEPROM position for storage split setting
#define BRIGHT      11 // EEPROM position for storage bright setting
#define KEY        129 // Random value between 0 and 256 to determine if the EPPROM has been written to
#define date __DATE__  // Compile date
#define time __TIME__  // Compile time

// Create general variables
int fix = 2,                 // Number of decimal digits in fix format
    state = STATE_WAIT,      // State for the number input state machine, default: wait for input
    modus = MODUS_SLEEP,     // Modus for the calculator, default:sleep
    set = SET_SCLOCK,        // Preferred setting for showing clock or count down, default: show clock
    sp = 0,                  // Stack pointer - points to the next free stack location
    pressed = 0,             // The key that was pressed
    numchar = 0,             // The number of characters in the input string
    digit = 0,               // The key that was pressed in use by store/recall and setting time and count-down
    digits,                  // Length of the complete string
    length,                  // Length of the showed value (is length-1 if a dot is present)
    pdot,                    // Position of the decimal dot
    pexp,                    // Position of the exponent letter E
    zeros,                   // Number of required trailing zeros
    mode,                    // mode of showing the results
                             // 1 x.x | Ey
                             // 2 x.x | y.y
                             // 3 xxx | x.x 
                             // 4   x.| x
    sign,                    // sign for the days of the count-down clock
    sec,                     // seconds for the count-down clock
    brightness,              // Brightness level 1-6
    pwm,                     // minmum brightness level for the LED's
    i,                       // temporary indices 
    j,
    k;

unsigned long int gosleep;    // Used for the sleep timer

const int keypad[2][5][4] ={{{40,1,2,3},{41,42,43,3},{44,45,46,5},{47,48,49,6},{7,8,9,10}},
                           {{11,12,13,14},{15,16,17,14},{19,20,21,22},{23,24,25,26},{27,28,29,30}}};

const int uitkeys[4] = {UIT1,UIT2,UIT3,UIT4};
const int inkeys[5] = {IN1,IN2,IN3,IN4,IN5};

bool lastkey = false,        // key must return to upper position before a new keypress may be detected
     shift = false,          // Shift key pressed
     changed = false,        // has the stack and/or input been changed
     error = false,          // not enough items on the stack to perform the operation
     overfl = false,         // digit overfl
     split = false,          // Use a split screen for integer (upper screen) and fractional (lower screen) part
     nofix = false;          // Don't round the number to a fix number of decimals

// Create string and char objects
char nr_c[STRING_LEN];       // the number that is being entered
String nr_s;                 // the number that is being entered 
char to_display[STRING_LEN]; // the number that is being displayed
String disp_s;               // the number that has to be displayed
char **endptr;               // not used but has to bedeclared

// Create float64 objects
float64_t stack[STACK_SIZE]; // the stack
float64_t expo[5];           // constant array of multiples of 10
float64_t memory[10];        // Calculator memory
float64_t x = fp64_sd(0.0),  // temporary x value of the stack
          y = fp64_sd(0.0),  // temporary y value of the stack
      d_num,                 // number that has to be showed
     mineen = fp64_sd(-1.0); // -1

// Create DateTime objects
DateTime nu, lol, hulp;
TimeSpan remain;

// Create display objects
TM1637TinyDisplay6 displayU(CUP, DUP); // define upper display
TM1637TinyDisplay6 displayL(CLO, DLO); // define lower display

uint8_t upper[6]; // buffer upper display
uint8_t lower[6]; // buffer lower display

const uint8_t zero[4] = {0x3F,0xBF,0x3F,0x3F}; // Count down zero
const uint8_t zerodot[4] = {0x3F,0xBF,0x3F,0xBF}; // Count down zero with dot
const uint8_t perm[6] = {0x73,0x79,0x31,0x33,0x27,0x00}; // Text PERM 

#if DUTCH
  const uint8_t  err[LOLEN] = {0x40,0x71,0x5C,0x1C,0x78,0x40}; // -fout- error message
  const uint8_t over[LOLEN] = {0x40,0x3F,0x3E,0x79,0x31,0x40}; // -over- message
  const uint8_t disp[LOLEN] = {0x40,0x5E,0x30,0x6D,0x73,0x40}; // -disp- error message
  const uint8_t get1[UPLEN] = {0x5E,0xDE,0x54,0xD4,0x0E,0x0E}; // dd.nn.jj
  const uint8_t get2[LOLEN] = {0x1C,0x9C,0x54,0xD4,0x3F,0x3F}; // uu.nn.00
#else
  const uint8_t  err[LOLEN] = {0x40,0x39,0x5F,0x38,0x39,0x40}; // -calc- error message
  const uint8_t over[LOLEN] = {0x40,0x3F,0x3E,0x79,0x31,0x40}; // -over- error message
  const uint8_t disp[LOLEN] = {0x40,0x5E,0x30,0x6D,0x73,0x40}; // -disp- error message
  const uint8_t get1[UPLEN] = {0x5E,0xDE,0x54,0xD4,0x6E,0x6E}; // dd.nn.yy - > change if required
  const uint8_t get2[LOLEN] = {0x74,0xF4,0x54,0xD4,0x3F,0x3F}; // hh.nn.00 
#endif

//const uint8_t table[UPLEN] = {0x78,0x5F,0x7C,0x38,0x79,0x40}; // table text 

const uint8_t dtoc[] = {       // character array with used ciphers
  0b01000000,    // -   0
  0b10000000,    // .   1
  0b00001000,    // _   2
  0b00111111,    // 0   3
  0b00000110,    // 1   4
  0b01011011,    // 2   5
  0b01001111,    // 3   6
  0b01100110,    // 4   7
  0b01101101,    // 5   8
  0b01111101,    // 6   9
  0b00000111,    // 7  10
  0b01111111,    // 8  11
  0b01101111,    // 9  12
  0b10111111,    // 0. 13
  0b10000110,    // 1. 14
  0b11011011,    // 2. 15
  0b11001111,    // 3. 16
  0b11100110,    // 4. 17
  0b11101101,    // 5. 18
  0b11111101,    // 6. 19
  0b10000111,    // 7. 20
  0b11111111,    // 8. 21
  0b11101111,    // 9. 22
  0b01111001,    // E  23
  0b00000000,    //    24
  0b00000010     // '  25 there are more digts
  };

// PROCEDURES ====================================================================================================================================

// Input procedure ----------------------------------------------------------------

int readkeys() { // Scan the keyboard
  int knop = 0,
      count = 0,
      temp = 0;  

  for (i = 0; i < 4; i++) { // activate columns/uitkeys
    digitalWrite(uitkeys[i], SELECT);
    for (j = 0; j < 5; j++) { // read rows/inkeys
      temp = 0;
      for (int k = 1; k <= DEBOUNCE; k++) temp += !digitalRead(inkeys[j]); // debounce
      temp = (temp > (DEBOUNCE/2)); // treshold reached
      count += temp; // count number of keypresses detected
      knop += (temp * keypad[shift][j][i]);
    }
    digitalWrite(uitkeys[i], DESELECT);
  }
  if (count != 1) { // reject two keys pressed
    knop = 0;
    count = 0;
  }

  if ((lastkey) && (count == 1)) knop = 0; // key must be released before a new press may be detected
  if ((!lastkey) && (count == 1)) lastkey = true; // key is pressed 
  if ((lastkey) && (count == 0)) lastkey = false; // key has been released 
  // process shift
  if (knop == 7) shift = true;
  if (knop == 27) shift = false;
  if ((knop > 10) && (knop < 31)) shift = false;
  #if TEST
    if (knop != 0) {
      Serial.print("Key:"); Serial.println(knop);
    }
  #endif
  return knop;
}

// Calculator procedures ----------------------------------------------------------------

void push(float64_t val) { // Push the stack with val
  if (sp == (STACK_SIZE)) {
    for (i = 0; i < STACK_SIZE-1; i++) stack[i] = stack[i+1]; // discard the upper stack value, so the stack will still be usable
    stack[sp-1] = val;
  } else {
    stack[sp++] = val;
  }
  state = STATE_WAIT;
  strcpy(nr_c, "+");
  numchar = 0;
  overfl = false;
  error = false;
}

float64_t pop() { // Pop the stack and return the value
  if (sp == 0) {
    error = true;
    return fp64_sd(0.0); 
  }
  return stack[--sp];
  overfl = false;
}

bool ready(int oper) { // Push the entered number on the stack and check voor the required oper number of operands

  if (state != STATE_WAIT) {
    push(fp64_strtod(nr_c, endptr)); // converts the entered string to a fp64 
  }
  error = (sp < oper);
  
  return !error;
}

void clear() { // Clear the memory
  for (i = 0; i <= 9; i++) memory[i] = fp64_sd(0); // clear memory
  memory[1] = fp64_sd(MEM1);
  memory[2] = fp64_sd(MEM2);
  memory[3] = MEM3;
  memory[4] = fp64_sd(MEM4);
  memory[5] = fp64_sd(MEM5);
  memory[6] = fp64_sd(MEM6);
  memory[7] = fp64_sd(MEM7);
  memory[8] = fp64_sd(MEM8);
  memory[9] = fp64_sd(MEM9);
  for (i = 0; i < STACK_SIZE; i++) stack[i] = fp64_sd(0); // clear stack
  sp = 0; 
  numchar = 0; 
  strcpy(nr_c, "+");
  overfl = false;
}

// Display procedures ----------------------------------------------------------------

void fixit(int point, bool todo) { // apply the fixed length and implicitly return to_display
 
  if (todo) {
    x = stack[point]; 
    if (!nofix) {
      x = fp64_mul(x, expo[fix]);
      x = fp64_round(x);
      x = fp64_div(x, expo[fix]);
    }
    disp_s = fp64_to_string(x, MAXLEN, MAXLEN-2); // convert to string
    pdot = disp_s.indexOf('.');
    if (pdot > 0) { // ad zero's until the fix length is reached
      zeros = 1 + fix - (disp_s.length() - pdot);
      if (zeros > 0) for (i = 0; i < zeros; i++) disp_s = disp_s + "0";
    }
  } else {
    disp_s = "";
    disp_s = nr_c; 
    if (nr_c[0] == '+') disp_s = disp_s.substring(1, disp_s.length()); // remove plus sign
    if (!overfl) disp_s = disp_s + CURSOR; // add cursor
  }
  // now disp_s contains the string that has to be displayed
      
  disp_s.toCharArray(to_display, disp_s.length()+1); // make a char array for display used in process
  pdot = disp_s.indexOf('.'); // position of the decimal point, -1 if not there
  pexp = disp_s.indexOf('E'); // position of the exponent, -1 if not there
  digits = disp_s.length(); 
  length = digits;            // length of the string
  if (pdot > 0) digits -= 1;  // mumber of 7-segment digits
  // now to_display contains the string that has to be displayed
}

uint8_t process(int pos) { // process a certain character in to_display and return the 7-segment byte
  int h;
  char a;  

  if (pos == -1) {
    a = ' '; h = 24; 
  } else { 
    a = to_display[pos];
    h = a - '-'; // convert char to array index
    if ((pdot > 0) && (pdot == pos+1)) { // DP detected?
      h += 10; // Add point
      i++;     // Increment the read pointer
    }
    if (a == CURSOR) h = 2; // display the underscore
    if (a == '+') h = 24;   // for the exponent, show space instead of +
  }

  #if TEST
    Serial.print(pos);  Serial.print("->"); Serial.print(j);
    Serial.print(",'"); 
    Serial.print(a); if (h > 12) Serial.print("."); 
    Serial.print("' | ");
  #endif
  j++; // increment the write pointer
  return dtoc[h];
};

// Date and time procedures ----------------------------------------------------------------

DateTime getit(DateTime n) { // Get and return a date and time

  int h, nu[6];

  nu[DAY] = n.day();
  nu[MONTH] = n.month();
  nu[YEAR] = n.year()-2000;
  nu[HOUR] = n.hour();
  nu[MINUTE] = n.minute();
  
  displayU.setSegments(get1); // write upper display yy.mm.dd
  displayL.setSegments(get2); // write lower display hh.mm.ss
  for(k = 0; k < 10; k++) {
    do digit = readkeys(); while ((digit < 40) || (digit > 49));
    h = ((k/2)+1.4); // calculate index
    if ((k % 2)) nu[h] = nu[h] + (digit-40); else nu[h] = (digit-40)*10;
    if ((k == 1) || (k == 3) || (k == 7) || (k == 9)){ 
      if (k < 6) displayU.setSegments(dtoc[digit-27], k);   // upper display odd with dp
      else       displayL.setSegments(dtoc[digit-27], k-6); // lower display odd with dp
    } else { // k is even
      
      if (k < 6) displayU.setSegments(dtoc[digit-37], k);   // upper display even
      else       displayL.setSegments(dtoc[digit-37], k-6); // lower display even
    }
  }
  changed = true;  // reset the display
  n = DateTime(nu[YEAR]+2000, nu[MONTH], nu[DAY], nu[HOUR], nu[MINUTE], 0);
  return n;
};

DateTime dstclock(DateTime n) { // Return the given (DST adjusted) date and time according to DST settings
  
  DateTime b, e;
  
  b = DateTime(n.year(), 3, 31, 3, 0, 0); b = DateTime(n.year(), 3, 31 - b.dayOfTheWeek(), 1, 0, 0);
  e = DateTime(n.year(), 10, 31, 3, 0, 0); e = DateTime(n.year(), 10, 31 - e.dayOfTheWeek(), 3, 0, 0);

  if (USEDST && (n > b) && (n < e)) n = n + TimeSpan(0,1,0,0); // adjust standard time if within summertime
  return n;
}; 

void setclock(DateTime n) { // if the clock is set during summertime then adjust the clock to standard time

  if (USEDST && (n != dstclock(n))) n = n - TimeSpan(0,1,0,0); // // if summertime then adjust to the standard time
  rtc.adjust(n); // Set the clock to standard time
}; 

void prepare(int pos, int num, bool up) { // write the clock item (year, month, day, hour, minute, second) in the display buffer
 
  k = (pos-1)*2;
  i = num/10; // left digit  
  j = num-(i*10); // right digit
  i = i + 3;
  j = j + 3; 
  if (pos != 3) j += 10; // add a decimal point
  if (up) {
     upper[k] = dtoc[i];
     upper[k+1] = dtoc[j]; 
   } else {
     if ((pos != 3) && (millis() / 1000) % 2) j -= 10; // blink time points
     lower[k] = dtoc[i];
     lower[k+1] = dtoc[j]; 
   }
}

#if TEST
  void printit() {
    nu = rtc.now();
    Serial.print("  Standard: ");Serial.print(nu.day()); Serial.print("-"); Serial.print(nu.month()); Serial.print("-"); Serial.print(nu.year()); Serial.print(" ");
    Serial.print(nu.hour()); Serial.print(":"); Serial.print(nu.minute()); Serial.print(":"); Serial.println(nu.second());
    nu = dstclock(rtc.now());
    Serial.print("    Actual: ");Serial.print(nu.day()); Serial.print("-"); Serial.print(nu.month()); Serial.print("-"); Serial.print(nu.year()); Serial.print(" ");
    Serial.print(nu.hour()); Serial.print(":"); Serial.print(nu.minute()); Serial.print(":"); Serial.println(nu.second());
    Serial.print("  Count to: ");Serial.print(lol.day()); Serial.print("-"); Serial.print(lol.month()); Serial.print("-"); Serial.print(lol.year()); Serial.print(" ");
    Serial.print(lol.hour()); Serial.print(":"); Serial.print(lol.minute()); Serial.print(":"); Serial.println(lol.second());
    if (lol > nu ) remain = lol - nu; else remain = nu - lol;
    sign = (lol > nu) ? -1 : 1;
    Serial.print("    Remain: ");Serial.print(remain.days()*sign); Serial.print(":"); Serial.print(remain.hours()); Serial.print(":"); Serial.print(remain.minutes()); Serial.print(":"); Serial.println(remain.seconds());
  }
#endif

// SETUP ==============================================================================================================================================================

void setup() {

  /* define I/O  */
  pinMode(UIT1, OUTPUT);
  pinMode(UIT2, OUTPUT);
  pinMode(UIT3, OUTPUT);
  pinMode(UIT4, OUTPUT);
  pinMode(IN1, INPUT_PULLUP);
  pinMode(IN2, INPUT_PULLUP);
  pinMode(IN3, INPUT_PULLUP);
  pinMode(IN4, INPUT_PULLUP);
  pinMode(IN5, INPUT_PULLUP);
  pinMode(HEART, OUTPUT);
  pinMode(LUP, OUTPUT);
  pinMode(LLO, OUTPUT);

  // initialise the rtc 
  rtc.begin();
  if (rtc.lostPower()) {
    setclock(DateTime(F(__DATE__), F(__TIME__)));
    nu = dstclock(rtc.now());
  } 
 
  // initialise the EEPROM 
  if (EEPROM.read(0) != KEY) { // if first time use of EEPROM then initialize
    EEPROM.update(0, KEY);    // EEPROM has been initialized
    nu = dstclock(rtc.now());
    EEPROM.update(YEAR, nu.year()-2000); // year
    EEPROM.update(MONTH, nu.month());    // month
    EEPROM.update(DAY, nu.day());        // day
    EEPROM.update(HOUR, nu.hour());      // hour
    EEPROM.update(MINUTE, nu.minute());  // minute
    EEPROM.update(SECOND, 0);            // seconds
    EEPROM.update(SET, set);             // Default: See set declaration
    EEPROM.update(FIX, 2);               // Number of decimals behind the decimal point
    EEPROM.update(NOFIX, 0);             // Default: off, e.g. round to fix
    EEPROM.update(SPLIT, 0);             // Use the splitscreen for showing results, default: off
    EEPROM.update(BRIGHT, 3);            // The set brightness
  }

  // read the EEPROM 
  lol = DateTime(EEPROM.read(YEAR) + 2000, EEPROM.read(MONTH), EEPROM.read(DAY), EEPROM.read(HOUR), EEPROM.read(MINUTE), EEPROM.read(SECOND));
  lol = dstclock(lol); // Adjust for DST settings
  set = EEPROM.read(SET); 
  fix = EEPROM.read(FIX); 
  nofix = EEPROM.read(NOFIX);
  split = EEPROM.read(SPLIT);
  brightness = EEPROM.read(BRIGHT);

  #if TEST
    // initialise the serial port
    Serial.begin(BAUD);
    Wire.begin();
    Serial.println(""); 
    Serial.println("===================== Begin ===================== ");
    Serial.println((String) "  Software: Retro RPN calculator");
    Serial.println((String) " Copyright: CC BY-NC-SA Jeroen Brinkman");
    Serial.println((String) "   Version: " + SOFTWARE);
    Serial.println((String) "  Compiled: " + date + " - " + time);
    printit();
    if (USEDST) Serial.println("       DST: On");
    else Serial.println("       DST: Off");
    Serial.println((String) "Brightness: " + brightness);
    Serial.println("------------------------------------------------");
  #endif

  // initialise the calculator
  state = STATE_WAIT;
  gosleep = millis() + SLEEP;
  zeros = 0;
  expo[0] = fp64_sd(1);
  expo[1] = fp64_sd(10);
  expo[2] = fp64_sd(100);
  expo[3] = fp64_sd(1000);
  expo[4] = fp64_sd(10000);
 
  clear();

  // initialise the displays
  for (i = 0; i < UPLEN; i++) upper[i] = 0x00; // clear upper display
  for (i = 0; i < LOLEN; i++) lower[i] = 0x00; // clear upper display
  displayU.begin();
  displayL.begin();

  // Process the brightess
  pwm = map(brightness,0,7,MINPWM,254);
  displayU.setBrightness(brightness);
  displayL.setBrightness(brightness + BRICOR);

  // Show the startscreen
  displayU.showString("RETRO-");
  displayL.showString("RPN jb");
  analogWrite(LUP, 0);
  analogWrite(LLO, 0);
}

// MAIN ==============================================================================================================================================================

void loop() {
  digitalWrite(HEART, (millis() / 1000) % 2); //1s heartbeat voor de Arduino LED
  nu = dstclock(rtc.now());  
  // scan the keyboard
  pressed = readkeys(); 
  changed = ((pressed != 0) && (pressed != 7));

  // process going to sleep
  if ((pressed != 0) || (modus == MODUS_SLEEP)) gosleep = millis() + SLEEP; // if key pressed or sleeping, reset sleep timer
  if (millis() > gosleep) { // Sleep timout ocurred
    EEPROM.update(SET, set);           // Store display settings
    EEPROM.update(FIX, fix);           // Store number of decimals behind the decimal point
    EEPROM.update(NOFIX, nofix);       // Store dont round to fix
    EEPROM.update(SPLIT, split);       // Store use the splitscreen for showing results
    EEPROM.update(BRIGHT, brightness); // Store brightness
    if (set == SET_PCLOCK) modus = MODUS_CLOCK;
    else if (set == SET_PCOUNT) modus = MODUS_COUNT;
    else modus = MODUS_SLEEP;
    if (modus == MODUS_SLEEP) { 
      displayU.clear();
      displayL.clear();
    }
  }

  // set the LED's
  if (modus == MODUS_CALC) analogWrite(LUP, pwm);
  else {
    analogWrite(LUP, 0);
    analogWrite(LLO, 0);
  }

  // process the awakening
  if ((changed) && (modus == MODUS_SLEEP)) { 
    if ((pressed > 39) && (pressed < 50)) {
      modus = MODUS_CALC;
      pressed = 99;
    } else { 
      if ((set == SET_SCLOCK) || (set == SET_PCLOCK)) modus = MODUS_CLOCK;
      else if ((set == SET_SCOUNT) || (set == SET_PCOUNT)) modus = MODUS_COUNT;
      changed = false;
    }
    displayU.setBrightness(brightness);
    displayL.setBrightness(brightness + BRICOR);
  }

  // process generic key presses
  if (changed) { 
    switch (pressed) {
    case 17: // Show clock / count down / calculator
      modus--;
      if (modus == MODUS_SLEEP) modus = MODUS_COUNT;
      if (modus == MODUS_CALC) pressed = 99;
      changed = false;
      break;
    case 21: // Brightness control 0 - 7
       brightness++;
       if (brightness == 8) brightness = 0;
       pwm = map(brightness,0,7,MINPWM,MAXPWM);
       displayU.setBrightness(brightness);
       displayL.setBrightness(brightness + BRICOR);
       #if TEST
         Serial.println((String) "Brightness: " + brightness);
       #endif
       changed = false;
      break;
    case 23: // Set clock date & time
      hulp = getit(nu); // enter date and time
      do digit = readkeys(); while(digit == 0); // catch the next key press
      if (digit == 3) { // only when the entered key is pressed
        setclock(hulp);
        #if TEST
          printit();
        #endif
      }
      changed = false;
      break;
    case 24: // Set count-down date & time
      hulp = getit(nu);  // enter date and time
      do digit = readkeys(); while(digit == 0); // catch the next key press
      if (digit == 3) { // only when the entered key is pressed
        lol = dstclock(hulp);
        EEPROM.update(YEAR, lol.year()-2000); // year
        EEPROM.update(MONTH, lol.month());    // month
        EEPROM.update(DAY, lol.day());        // day
        EEPROM.update(HOUR, lol.hour());      // hour
        EEPROM.update(MINUTE, lol.minute());  // minute
        EEPROM.update(SECOND, lol.second());  // seconds
        #if TEST
          printit();
        #endif
      }
      changed = false;
      break;
    case 25: // Display item: off | date&time | count-down
      set --; 
      if (set == 0) set = SET_PCOUNT;
      switch (set) {
      case SET_OFF:
        displayU.showString("Clock ");
        displayL.showString("Off   ");
        break;
      case SET_PCLOCK:
        displayU.setSegments(perm);
        displayL.showString("Clock ");
        break;
      case SET_PCOUNT:
        displayU.setSegments(perm);
        displayL.showString("Count ");
        break;
      case SET_SCLOCK:
        displayU.showString("SLEEP ");
        displayL.showString("Clock ");
        break;
      case SET_SCOUNT:
        displayU.showString("SLEEP ");
        displayL.showString("Count ");
        break;
      }
      delay(SEEDELAY);
      changed = false;
      break;
    }
  }

// Process clock and count down
  if (changed && ((modus == MODUS_CLOCK) || (modus == MODUS_COUNT))) { //switch from displaying to calculator
    modus = MODUS_CALC;
    pressed = 99;
  }

  if (modus == MODUS_CLOCK) { // display date & time
    prepare(DD, nu.day(), true);
    prepare(MM, nu.month(), true);
    prepare(YY, nu.year()-2000, true);
    prepare(1, nu.hour(), false);
    prepare(2, nu.minute(), false);
    prepare(3, nu.second(), false);
    displayU.setSegments(upper); // write upper display
    if (shift) lower[LOLEN-1] = (lower[LOLEN-1] | SEG_DP); // display shift dot
    displayL.setSegments(lower); // write lower display
  }

  if (modus == MODUS_COUNT) { // display count-down & count-up
    if (lol > nu) {
      remain = lol - nu;
      sign = -1;
    } else {
      remain = nu - lol;
      sign = 1;
    }
    if (sign == -1) { 
      sec = 60-nu.second();
      if (sec == 60) sec = 0;
    } else sec = nu.second();
    if ((sign == -1) && (remain.days() == 0)) displayU.showString("    -0");
    else displayU.showNumber(remain.days()*sign, false); // write upper display
    prepare(1, remain.hours(), false);
    prepare(2, remain.minutes(), false);
    prepare(3, sec, false);
    if (shift) lower[LOLEN-1] = (lower[LOLEN-1] | SEG_DP); // display shift dot
    displayL.setSegments(lower); // write lower display
  }
  
  // process calculator related key presses  ---------------------------------------------------------------------

  if (changed && (modus == MODUS_CALC)) { 
    switch (pressed) {
    case 0: // Nothing pressed so NOP
      // statements
      break;
    case 1: // .
      if (state == STATE_WAIT) {  // start with a new number
        strcpy(nr_c, "+0.");
        numchar = 1;
      }
      if (state == STATE_INT) {
       strcat(nr_c, ".");
      }
      state = STATE_FRAC;
      break;
    case 2: // 00
      if (state != STATE_WAIT) { // don't start with two zeroes
        if (numchar + 2 >= MAXLEN) overfl = true;
        else {
          strcat(nr_c, "0"); 
          strcat(nr_c, "0");
          numchar += 2;
        }
      }
      break;
    case 3: // Enter
      if (state == STATE_WAIT) {  // Duplicate the number on the stack
        x = pop();
        if (!error) {
          push(x);
          push(x);
        }
      } else {
        push(fp64_strtod(nr_c, endptr)); // converts a string to a fp64 
      }
      delay(WAIT);
      break;
    case 5: // +
      if (ready(2)) {
        x = pop();
        y = pop();
        push(fp64_add(y, x));
      }
      break;
    case 6: // -
      if (ready(2)) {
        x = pop();
        y = pop();
        push(fp64_sub(y, x));
      }
      break;
    case 8: // /
      if (ready(2)) {
        if (stack[sp-1] == 0) error = true;
        else {
          x = pop();
          y = pop();
          push(fp64_div(y, x));
        }
      }
      break;
    case 9: // *
      if (ready(2)) {
        x = pop();
        y = pop();
        push(fp64_mul(y, x));
      }
      break;
    case 10: // Backspace
      nr_s = nr_c;
      if (nr_c[nr_s.length()-1] == '.') state = STATE_INT; 
      if (nr_s.length() > 1) nr_c[nr_s.length()-1] = '\0';
      numchar -= 1;
      overfl = false;
      break;
    case 11: // Fix
      do digit = readkeys(); while(digit == 0);
      if ((digit > 39) && (digit < 46)) fix = digit - 40;
      switch (digit) {
        case 46: // 6 pressed
          displayU.showString("RETRO-");
          displayL.showString("RPN jb");
          break;
        case 47: // 7 pressed
          displayL.showString(SOFTWARE);
          delay(SEEDELAY);
          break;
        case 48: // 8 pressed
          displayU.showString("[c]'24");
          displayL.showString("JEROEN");
          break;
        case 49: // 9 pressed
          // NOP
          break;
        default: // 0 - 4 pressed
          displayL.showString("FIX   ");
          displayL.showNumber(fix, false, 1, 4);
      }
      delay(SEEDELAY);
      nofix = false;
      break;
    case 12: // Clear all
      clear();
      break;
    case 13: // Swap
      if (ready(2)) {
        x = pop();
        y = pop();
        push(x);
        push(y);
      }
      break;
    case 14: // Rotup (Rotate stack)
      if (ready(2)) {
        x = stack[sp-1];
        for (int i = sp-1; i > 0; i--) stack[i] = stack[i-1];
        stack[0] = x;
      }
      break;
      delay(WAIT);
      break;
    case 15: // Togggle nofix view
      nofix = !nofix;
      if (nofix) {
        displayL.showString("NOFIX ");
      } else { 
        displayL.showString("FIX   ");
        displayL.showNumber(fix, false, 1, 4);
      }
      delay(SEEDELAY);
      break;
    case 16: // Toggle split view
      split = !split;
      if (split) displayL.showString("SPLIT ");
      else displayL.showString("NOSPLT");
      delay(SEEDELAY);
      break;
    case 19: // Store
      do digit = readkeys(); while(digit == 0);
      if (state == STATE_WAIT) {
        if (sp == 0) error = true; else x = stack[sp-1]; // retreive the value to store from the stack
      } else x = fp64_strtod(nr_c, endptr); // retreive the value to store from the actual input
      if ((state == STATE_INT) && (digits == 1)) error = true; // check if nothing is entered
      else { 
        if ((digit > 39) && (digit < 50)) memory[digit - 40] = x; // store value
      }
      break;
    case 20: // Recall
      if (state != STATE_WAIT) push(fp64_strtod(nr_c, endptr)); // push the value on the stack before recall
      do digit = readkeys(); while(digit == 0);
      if ((digit > 39)&& (digit < 50)) push(memory[digit - 40]);
      break;
    case 22: // y power x
      if (ready(2)) {
        x = pop();
        y = pop();
        push(fp64_pow(y, x));
      }
      break;
    case 26: // -x
      if ((sp == 0) && (state == STATE_WAIT)) state = STATE_INT;  // start with a new number
      if (state == STATE_WAIT) { // change number on the stack
        x = fp64_mul(pop(), mineen);
        push(x);
      } else {
        numchar = (nr_c[0] == '+') ? (numchar + 1) : (numchar - 1);
        if (nr_c[0] == '+') nr_c[0] = '-'; else nr_c[0] = '+';
      }
      break;
    case 27: // Shift
      // no operation
      break;
    case 28: // 1/x
      if (ready(1)) {
        if (stack[sp-1] == 0) error = true;
        else {
          x = pop(); 
          push(fp64_inverse(x));
        }
      }
      break;
    case 29: // %
      if (ready(2)) {
        x = fp64_div(pop(), expo[2]);
        y = pop();
        push(fp64_mul(y, x));
      }
      break;
    case 30: // clear
      if (state == STATE_WAIT) {
        if (sp != 0) pop();
      } else {
        state = STATE_WAIT;
        strcpy(nr_c, "+");
        numchar = 0;
        overfl = false;
        error = false;
      }
      break;
    default: // process numbers
      if (state == STATE_WAIT) {  // start with a new number
        state = STATE_INT;        // and the digits entered make up the integer part of the number
      }
      if (numchar <= MAXLEN) {
        if (pressed == 40)      strcat(nr_c, "0");
        else if (pressed == 41) strcat(nr_c, "1");
        else if (pressed == 42) strcat(nr_c, "2");
        else if (pressed == 43) strcat(nr_c, "3");
        else if (pressed == 44) strcat(nr_c, "4");
        else if (pressed == 45) strcat(nr_c, "5");
        else if (pressed == 46) strcat(nr_c, "6");
        else if (pressed == 47) strcat(nr_c, "7");
        else if (pressed == 48) strcat(nr_c, "8");
        else if (pressed == 49) strcat(nr_c, "9");
        numchar += 1; error = false;
      } 
      if (numchar == MAXLEN) overfl = true;
      break;
    }
  }
  // Display the value(s)  ---------------------------------------------------------------------
  if ((pressed != 0) && (modus == MODUS_CALC)) {
    fixit(sp-1, state == STATE_WAIT); // Convert the value in a string

    // initialize variables 
    mode = 0;
    for(i = 0; i < UPLEN; i++) upper[i] = dtoc[24]; // clear array upper display
    for(i = 0; i < LOLEN; i++) lower[i] = dtoc[24]; // clear array lower display

    // determine display mode 
    if (pexp > 0)             mode = SHOW_ENG;   //  x.x | Ey
    else if (digits <= UPLEN) mode = SHOW_TWO;   //  x.x | y.y
    else                      mode = SHOW_EXT;   //  xxxxxx | x.x
    
    if ((fp64_trunc(stack[sp-1]) != stack[sp-1]) && (mode > SHOW_ENG) && split && (state == STATE_WAIT)) mode = SHOW_SPLIT; // when splitscreen is selected and applicable

    // execute display mode 
    #if TEST
      Serial.print("Upper parse:"); 
    #endif

    switch (mode) {
    case SHOW_ENG: // x.x | Ey
      if (pexp > UPLEN) {
        for(i = 0; i < LOLEN; i++) lower[i] = disp[i]; 
      } else {
        j = 0; // set write pointer to 0
        for (i = 0; i < (UPLEN + 1 - pexp); i++) upper[i] = process(-1);// write spaces
        for(i = 0; i < pexp; i++)                upper[j] = process(i); // write digits
        lower[0] = dtoc[23]; // write E
        j = 1; // set write pointer to 1
        for (i = pexp+1; i < length; i++)        lower[j] = process(i); // write spaces
      }
      analogWrite(LLO, 0);
      break;
    case SHOW_TWO: // x.x | y.y
      j = 0; // set write pointer to 0
      if (state == STATE_WAIT) 
        for (i = 0; i < (UPLEN - digits); i++) upper[i] = process(-1);// write spaces
      for(i = 0; i < length; i++)              upper[j] = process(i); // write digits 
 
      // show second value on the lower display if applicable
      k = sp; 
      if (state != STATE_WAIT) k += 1; // there is a value being entered
      if (k > 1) { // there are at least two values, show second value
        analogWrite(LLO, pwm);
        #if TEST
          Serial.println(" ");
          Serial.print("Lower parse:");
        #endif
        if (state == STATE_WAIT) fixit(sp-2, STACK); else fixit(sp-1, STACK); // prepare the second value for display
        k = digits;
        if (digits > LOLEN) k = LOLEN; // if number of digits larger than display size
        j = 0; // set write pointer to 0
        if ((pdot > 0) && (pdot < LOLEN)) k += 1; // add one if dp available
        for (i = 0; i < (LOLEN - digits); i++) lower[i] = process(-1);// write spaces
        for (i = 0; i < k; i++)                lower[j] = process(i); // write digits 
        if (digits > LOLEN) lower[LOLEN-1] = dtoc[25]; // if too many characters, show continue character
      }  else analogWrite(LLO, 0);
      break;
    case SHOW_EXT: //  xxxxxx | x.x
      j = 0; // set write pointer to 0
      k = (pdot < 0) ? UPLEN : UPLEN+1; // Deal with difference between length an digits
      for(i = 0; i < k; i++)               upper[j] = process(i); // write digits 
      j = 0; // set write pointer to 0
      for(i = k; i < length; i++)          lower[j] = process(i); // write digits 
      analogWrite(LLO, 0);
      break;
    case SHOW_SPLIT: //  xxx.| xx
      j = 0; // set write pointer to 0
      for (i = 0; i < (UPLEN - pdot); i++) upper[i] = process(-1);// write spaces
      for(i = 0; i < pdot; i++)            upper[j] = process(i); // write digits before dot
      j = 0; // set write pointer to 0
      for(i = pdot+1; i < length+1; i++)   lower[j] = process(i); // write digits after dot
      analogWrite(LLO, 0);
      break;
    }

    if (overfl) for(i = 0; i < LOLEN; i++) lower[i] = over[i]; // write over in lower window
    if (error) for(i = 0; i < LOLEN; i++) lower[i] = err[i]; // write error in lower window
    displayU.setSegments(upper); // write upper display
    if (shift) lower[LOLEN-1] = (lower[LOLEN-1] | SEG_DP); // display shift dot
    displayL.setSegments(lower); // write lower display
  }

  // debugging info ---------------------------------------------------------------------
  #if TEST
    if (pressed != 0) {
      Serial.println("");
      Serial.print("pressed:"); Serial.println(pressed);

      Serial.print("modus:"); Serial.print(modus);
      if (modus == MODUS_SLEEP)      Serial.println(" Sleep");
      else if (modus == MODUS_CLOCK) Serial.println(" Clock");
      else if (modus == MODUS_COUNT) Serial.println(" Count");
      else if (modus == MODUS_CALC)  Serial.println(" Calc");

      Serial.print("set:"); Serial.print(set);
      if (set == SET_OFF)          Serial.println(" Off");
      else if (set == SET_PCLOCK)  Serial.println(" Show time permanent");
      else if (set == SET_PCOUNT)  Serial.println(" Show count down permanent");
      else if (set == SET_SCLOCK)  Serial.println(" Show time and go to sleep");
      else if (set == SET_SCOUNT)  Serial.println(" Show count down and go to sleep");

      if (modus == MODUS_CALC) {
        Serial.println("");
        Serial.println("CALCULATOR");
        Serial.print("fix:"); Serial.println(fix);
        Serial.print("nofix:"); Serial.println(nofix);
        Serial.print("split:"); Serial.println(split);
        Serial.print("state:"); Serial.print(state);
        if (state == STATE_WAIT)      Serial.println(" Wait");
        else if (state == STATE_INT)  Serial.println(" Int");
        else if (state == STATE_FRAC) Serial.println(" Frac");

        Serial.print("display mode:"); Serial.print(mode);
        if (mode == SHOW_ENG)        Serial.println(" Engineering display");
        else if (mode == SHOW_TWO)   Serial.println(" Two diplays");
        else if (mode == SHOW_EXT)   Serial.println(" Extended display");
        else if (mode == SHOW_SPLIT) Serial.println(" Split display");


        Serial.print("numchar:"); Serial.println(numchar);
        Serial.print("length:"); Serial.println(length);
        Serial.print("pdot:"); Serial.println(pdot);
        Serial.print("pexp:"); Serial.println(pexp);
        Serial.print("zero's:"); Serial.println(zeros);
        Serial.print("digits:"); Serial.println(digits);
        Serial.print("disp:"); Serial.println(disp_s);
        Serial.print("sp:"); Serial.println(sp);
        Serial.println("------------------------------------------------");
        Serial.print("entered:"); Serial.println(nr_c);
        Serial.print("Stack ");     
        for (i = sp-1; i >= 0; i--) {
          if (i < 0) Serial.println("0");
          else {
            char *t = fp64_to_string(stack[i], MAXLEN, MAXLEN-2);
            Serial.print(i); Serial.print(": ");
            Serial.println(t);
            if (i != 0) Serial.print("      ");
          }
        }
      }
      Serial.println("------------------------------------------------");
    }
  #endif
}